#include "utils.h"

#include <Wire.h>
#include <LiquidCrystal_I2C.h>

// Set the LCD address to 0x27 for a 16 chars and 2 line display
LiquidCrystal_I2C lcd(0x27, 16, 2);

int presetType = PRESET_CONTROL;
int presetNumber = 0;
int channel = 0;
const int NUM_PRESETS = 7;
const float CHANGE_DELAY = 300;
float changeDelay = 0;
char lcdText[2][256];

int buttonState[NUM_BUTTONS] = {0};
bool buttonLastState[NUM_BUTTONS] = {0};
byte buttonControlState[NUM_BUTTONS] = {0};
byte ledStates[2] = {0, 0};
bool buttonPressed;

enum LoopModes {
    LOOP_MODE_MAIN,
    LOOP_MODE_IDLE,
    LOOP_MODE_CHANGE
};

int loopMode = LOOP_MODE_MAIN;
bool loopModeChanged = true;

void updateLcd() {
  sprintf(lcdText[0], "%s %d | CH %d", presetText[presetType], presetNumber, channel);
  if (loopMode == LOOP_MODE_CHANGE) {
      sprintf(lcdText[1], "CHANGE");
  }
  lcd.clear();
  lcd.setCursor(0, 0);
  lcd.print(lcdText[0]);
  lcd.setCursor(0, 1);
  lcd.print(lcdText[1]);
}

void setup() {
    ledStates[0] = ledStates[1] = 0xff;
    memset(buttonLastState, 0, NUM_BUTTONS);
    memset(buttonControlState, 0, NUM_BUTTONS);

    // Multiplexer
    pinMode(MUX_S0, OUTPUT);
    pinMode(MUX_S1, OUTPUT);
    pinMode(MUX_S2, OUTPUT);
    pinMode(MUX_S3, OUTPUT);
    pinMode(MUX_SIG, INPUT_PULLUP);

    // Shift register
    pinMode(SR_RCLK, OUTPUT);
    pinMode(SR_SRCLK, OUTPUT);
    pinMode(SR_SER, OUTPUT);

    digitalWrite(MUX_S0, LOW);
    digitalWrite(MUX_S1, LOW);
    digitalWrite(MUX_S2, LOW);
    digitalWrite(MUX_S3, LOW);

    Serial.begin(115200);

    // initialize the LCD
    lcd.begin();

    // Turn on the blacklight and print a message.
    lcd.backlight();

    updateLcd();
    updateShiftRegister(ledStates);
}

void updateButtons() {
    buttonPressed = false;
    for (int button = 0; button < NUM_BUTTONS; button++) {
        buttonLastState[button] = buttonState[button];
        buttonState[button] = !muxRead(button);
        buttonPressed |= buttonState[button];
  
        if (buttonState[button] != buttonLastState[button]) {
            setLed(button, ledStates, !buttonState[button]);
        }
    }
}

void loopIdle() {
    static int pos = 0;
    static float loopDelay = 0;
    static const float loopDelayMs = 500, effectDelayMs = 60;

    updateButtons();
    loopModeChanged = false;
    if (millis() - loopDelay > loopDelayMs) {
        loopDelay = millis();
        //ledStates[0] = ledStates[1] = pos % 2 ? 0xff : 0;
        ledStates[0] = ledStates[1] = 0xff;
        //setLed(rand() % NUM_BUTTONS, ledStates, 0);
        //updateShiftRegister(ledStates);
        setLed(pos, ledStates, 0);
        //setLed(pos + NUM_BUTTONS / 2, ledStates, 0);
        pos = (pos + 1) % (NUM_BUTTONS);
    }

    if (buttonPressed) {
        ledStates[0] = ledStates[1] = 0xff;
        updateShiftRegister(ledStates);
        loopMode = LOOP_MODE_MAIN;
        loopModeChanged = true;
    }
}

void loopChange() {
    updateButtons();

    if (loopModeChanged) {
        changeDelay = millis();
        updateLcd();
    }

    loopModeChanged = false;
    
    if (millis() - changeDelay >= CHANGE_DELAY) {
        if (buttonState[BUTTON_CONTROL]) {
            loopMode = LOOP_MODE_MAIN;
            loopModeChanged = true;
            updateButtons();
            return;
        }
      
        bool changed = false;
        if (buttonState[BUTTON_CONTROL_PRESET_TYPE]) {
            presetNumber = (presetNumber + 1) % NUM_PRESETS;
            changed = true;
        } else if (buttonState[BUTTON_CONTROL_PRESET_NO]) {
            presetType = (presetType + 1) % NUM_PRESET_TYPES;
            changed = true;
        } else if (buttonState[BUTTON_CONTROL_CHANNEL_PLUS]) {
            channel = (channel + 1) % NUM_CHANNELS;
            changed = true;
        } else if (buttonState[BUTTON_CONTROL_CHANNEL_MINUS]) {
            channel--;
            if (channel < 0) {
                channel = NUM_CHANNELS;
            }
            changed = true;
        }

        if (changed) {
            changeDelay = millis();
            updateLcd();
        }
    }
}

void loopMain() {
  static bool changedText = false;
  static char buttonsText[256] = "", aux[256];
  static const float idleTimeMs = 10000;
  static float idleTime = millis();
  static float changeTime = 0;
  
  buttonPressed = false;
  changedText = false;
  buttonsText[0] = 0;

  if (loopModeChanged) {
      idleTime = millis();
      changeTime = millis();
  } else {
      // Update button states
      updateButtons();
  }

  loopModeChanged = false;

  // There will be a change of preset type, number or channel
  if (buttonState[BUTTON_CONTROL] && millis() - changeTime >= CHANGE_DELAY) {
      loopMode = LOOP_MODE_CHANGE;
      loopModeChanged = true;
  } else {
      for (int button = 0; button < NUM_BUTTONS; button++) {
          if (button == BUTTON_CONTROL) {
              continue;
          }

          // The button was pushed
          if (buttonState[button]) {
              if (presetType == PRESET_CONTROL) {
                  sprintf(aux, "CC%d ", getControl(presetNumber, button));
                  strcat(buttonsText, aux);
              } else {
                  getNoteName(presetNumber, button, aux);
                  strcat(buttonsText, aux);
                  strcat(buttonsText, " ");
               }

              // It's a state change
              if (buttonState[button] != buttonLastState[button]) {
                 changedText = true;
                 if (presetType == PRESET_CONTROL) {
                    buttonControlState[button] = !buttonControlState[button];
                    controlChange(channel, getControl(presetNumber, button), buttonControlState[button] ? 127 : 0);
                 } else {
                    noteOn(channel, getNote(presetNumber, button), 127);
                 }
              }
          } else {
              // It's a state change
              if (buttonState[button] != buttonLastState[button]) {
                  changedText = true;
                  if (presetType != PRESET_CONTROL) {
                      noteOff(channel, getNote(presetNumber, button), 0);
                  } else {
                      buttonControlState[button] = !buttonControlState[button];
                      controlChange(channel, getControl(presetNumber, button), buttonControlState[button] ? 127 : 0);
                  }
              }
            }
      }

      MidiUSB.flush();
  }

  // If the text changed then update the LCD
  if (changedText || strcmp(buttonsText, lcdText[1]) != 0) {
      strcpy(lcdText[1], buttonsText);
      updateLcd();
  }

  // Check time for idle mode
  if (!buttonPressed && millis() - idleTime > idleTimeMs) {
      loopMode = LOOP_MODE_IDLE;
  } else if (buttonPressed) {
      idleTime = millis();
  }
}

void loop() {
    if (loopMode == LOOP_MODE_MAIN) {
        loopMain();
        return;
    }

    if (loopMode == LOOP_MODE_IDLE) {
        loopIdle();
        return;
    }
    
    if (loopMode == LOOP_MODE_CHANGE) {
        Serial.println("CHANGE");
        loopChange();
        return;
    }
}
